/*-*- c++ -*-********************************************************
 * MEditCtrl class: API for message display and edit controls     *
 *                                                                  *
 * (C) 2000 by Karsten Ballder (ballueder@gmx.net)                 *
 *                                                                  *
 * $Id: MEditCtrl.h,v 1.3 2000/08/07 19:05:17 KB Exp $
 *
 *******************************************************************/
#ifndef MEDITCTRL_H
#define MEDITCTRL_H

#ifdef __GNUG__
#   pragma interface "MEditCtrl.h"
#endif

#include "Mcommon.h"
#include "MObject.h"
#include "strutil.h"

#include <gui/wxllist.h> // CoordType

/** A small helper class holding some data.
    MimeContent holds the following:
    - a MIME type specification string
    - a reference number which is used by whoever creates it
      (called partno as it usually refers to a message part)
*/
class MimeContent : public MObject
{
public:
   /// create a text object
   MimeContent(const String &text, UIdType partno)
      {
         Create("text/plain",text, partno);
      }
   /// create a text object of a non-standard type
   MimeContent(const String &type, const String &text, UIdType partno)
      {
         Create(type, text, partno);
      }
   /** create a data object with some raw, non text data
       Takes control of the data pointer
   */
   MimeContent(const String &type, void * data, size_t len, UIdType partno)
      {
         m_Data = data;
         m_Len = len;
         m_PartNo = partno;
         SetType(type);
      }
   /// return the mime type
   const String & GetType(void) const
      { MOcheck(); return m_Type; }
   const wxString & GetText(void) const
      { MOcheck(); return m_Text; }
   /// return the part number
   UIdType GetPartNumber() const { MOcheck(); return m_PartNo; }

   const String &GetFileName(void) const { return m_FileName; }

   const void *GetData(void) const
      {
         MOcheck();
         if(m_Data)
            return m_Data;
         else
            return m_Text.c_str();
      }
   size_t GetSize(void) const
      {
         MOcheck();
         if(m_Data)
            return m_Len;
         else
            return m_Text.Length();
      }
   void SetFileName(const String &fn)
      {
         MOcheck();
         m_FileName = fn;
      }

   void SetType(const String &t)
      {
         MOcheck();
         m_Type = t;
         strutil_tolower(m_Type);
      }
   void SetDisposition(const String &d)
      {
         MOcheck();
         m_Disposition = d;
         strutil_tolower(m_Disposition);
      }
   const String &GetDisposition(void) const
      {
         return m_Disposition;
      }

   ~MimeContent()
      {
         MOcheck();
         FreeData();
      }
   void MakeDataCopy(void)
      {
         MOcheck();
         CHECK_RET(m_Data != NULL && m_Len > 0, "attempt to copy NULL data");
         void *newdata = malloc(m_Len);
         memcpy(newdata, m_Data, m_Len);
         m_Data = newdata;
      }
protected:
   void FreeData()
      {
         MOcheck();
         if(m_Data) free(m_Data);
         m_Data = NULL;
         m_Len = 0;
      }
   void Create(const String &type, const String &text, UIdType partno)
      {
         m_Data = NULL;
         m_Len = 0;
         m_Type = type;
         m_Text = text;
         m_PartNo = partno;
      }
private:
   String m_Type;
   UIdType m_PartNo;
   String m_FileName;
   String m_Disposition;
   String m_Text;
   void *m_Data;
   size_t m_Len;
   /// not implemented:
   MimeContent(const MimeContent &);
   MimeContent & operator=(const MimeContent &);
};

/// Mini-class implementing an XFace via MimeContent
class MimeContentXFace : public MimeContent
{
public:
   MimeContentXFace(const String &data)
      : MimeContent ("application/x-m-mc-xface", data, UID_ILLEGAL)
      {}
};

class FontStyle : public MObject
{
public:
   FontStyle(int family = wxROMAN,
             int size=12,
             int style=wxNORMAL,
             int weight=wxNORMAL,
             int underline=0,
             String fg="black",
             String bg="white")
      {
         m_Family = family;
         m_Size = size;
         m_Style = style;
         m_Weight = weight;
         m_Underline = underline;
         m_ColourFG = fg;
         m_ColourBG = bg;
      }
   int m_Family, m_Size, m_Style, m_Weight;
   bool m_Underline;
   String m_ColourFG, m_ColourBG;
};

/** This class interfaces with the edit control for callbacks. */
class MEditCtrlCbHandler : public MObject
{
public:
   /**@name Callbacks for editor control */
   //@{
   /// when clicked on mimepart
   virtual void OnMouseLeft(UIdType partno) = 0;
   /// when right-clicked on mimepart
   virtual void OnMouseRight(UIdType partno) = 0;
   /// when clicked on url
   virtual void OnUrl(const wxString &url) = 0;
   //@}
   virtual ~MEditCtrlCbHandler() {}
};

#define MEC_WXLAYOUT 1
//#define MEC_WXEMBED  2

class MEditCtrl : public MObject
{
public:
   /** Create an MEditCtrl with a pointer to a callback handler for
       interfacing and a parent window. Supported types:
       MEC_WXLAYOUT
   */
   static MEditCtrl * Create(int type,
                             MEditCtrlCbHandler *cbh, 
                             wxWindow *parent,
                             Profile *prof = NULL);

   /// Tell it to read the configuration settings
   virtual void ReadProfile(Profile *prof) = 0;
   /// clear the control
   virtual void Clear(void) = 0;

   /** Redraws the window. */
   virtual void RequestUpdate(void) = 0;

   /** Enable redrawing while we add content */
   virtual void EnableAutoUpdate(bool enable = TRUE) = 0;

   /** Set dirty flag */
   virtual void SetDirty(bool dirty = TRUE) = 0;
   virtual bool GetDirty(void) const = 0;

   /** Content can be inserted via three different methods:

       + Append(const MimeContent &)
         Insert arbitrary embedded/attached MIME content

       + Apppend(const String &txt)
         Insert text in text/plain

       + NewLine()
         Add a new-line.
   */
   /** This inserts some mime content of a non-text type into the
       control. 
   */
   virtual void Append(const MimeContent &mc) = 0;

   /** Bog-standard text-append to the control. Text is treated as
       "text/plain" */
   virtual void Append(const String &txt) = 0;

   virtual void NewLine(void) = 0;
   
   virtual bool Find(const wxString &needle,
                     wxPoint * fromWhere = NULL,
                     const wxString &configPath = "MsgViewFindString")
      = 0;

   /**@name Clipboard interaction */
   //@{
   /// Pastes text from clipboard.
   virtual void Paste(bool privateFormat = FALSE,
                      bool usePrimarySelection = FALSE) = 0;
   /** Copies selection to clipboard.
       @param invalidate used internally, see wxllist.h for details
   */
   virtual bool Copy(bool invalidate = true,
                     bool privateFormat = FALSE, bool primary = FALSE) = 0;
   /// Copies selection to clipboard and deletes it.
   virtual bool Cut(bool privateFormat = FALSE,
                    bool usePrimary = FALSE) = 0;
   ///    Enable or disable editing, i.e. processing of keystrokes.
   virtual void SetEditable(bool toggle) = 0;
   //@}

   /** Prints the currently displayed message.
       @param interactive if TRUE, ask for user input
       return TRUE on success
   */
   virtual bool Print(bool interactive = TRUE) = 0;
   
   /// print-previews the currently displayed message
   virtual void PrintPreview(void) = 0;
   /// moves the cursor to the new position
   virtual void MoveCursorTo(int x, int y) const = 0;

   virtual void PageUp(void) const = 0;
   virtual void PageDown(void) const = 0;
   
   /**@name GUI matters */
   //@{   
   /// adjust the window size
   virtual void SetSize(int x, int y) = 0;
   virtual void SetFontStyle(const FontStyle &fs) = 0;
   virtual void SetFocusFollowMode(bool enable) = 0;
   /** Sets the wrap margin.
       @param margin set this to 0 to disable it
   */
   virtual void SetWrapMargin(CoordType margin) = 0;
   /** Tell window to update a wxStatusBar with UserData labels and
       cursor positions.
       @param bar wxStatusBar pointer
       @param labelfield field to use in statusbar for URLs/userdata labels, or -1 to disable
       @param cursorfield field to use for cursor position, or -1 to disable
   */
   virtual void SetStatusBar(class wxStatusBar *bar,
                       int labelfield = -1,
                       int cursorfield = -1) = 0;

   /** Sets cursor visibility, visible=1, invisible=0,
       visible-on-demand=-1, to hide it until moved.
       @param visibility -1,0 or 1
       @return the old visibility
   */
   virtual void SetCursorVisibility(int visibility = -1) = 0;
   /**@name Formatting options */
   //@{
   /// set font family
   virtual void SetFontFamily(int family) = 0;
   /// set font size
   virtual void SetFontSize(int size) = 0;
   /// set font style
   virtual void SetFontStyle(int style)  = 0;
   /// set font weight
   virtual void SetFontWeight(int weight) = 0;
   /// toggle underline flag
   virtual void SetFontUnderline(bool ul) = 0;
   /// set font colours by colour
   virtual void SetFontColour(wxString fg, wxString bg) = 0;
   virtual wxPoint GetClickPosition(void) const = 0;
   //@}
   virtual ~MEditCtrl() {}
private:
};
#endif
