//////////////////////////////////////////////////////////////////////////////
// Project:     Mahogany - cross platform e-mail GUI client
// File name:   MailFolderCC.h: declaration of MailFolderCC class
// Purpose:     handling of mail folders with c-client lib
// Author:      Karsten Ballder
// Modified by:
// Created:     1997
// CVS-ID:      $Id: MailFolderCC.h,v 1.200 2001/12/23 16:50:31 vadz Exp $
// Copyright:   (C) 1997-2001 by Karsten Ballder (ballueder@gmx.net)
// Licence:     M license
///////////////////////////////////////////////////////////////////////////////

#ifndef MAILFOLDERCC_H
#define MAILFOLDERCC_H

#ifdef __GNUG__
#   pragma interface "MailFolderCC.h"
#endif

#ifndef   USE_PCH
#  include "kbList.h"

#  include "Mcclient.h"
#endif

#include "MThread.h"

#include "MFolder.h"

#include "MailFolderCmn.h"
#include "FolderView.h"

#include <wx/fontenc.h>    // enum wxFontEncoding can't be fwd declared

// fwd decls
class ASMailFolder;
class MailFolderCC;

// ----------------------------------------------------------------------------
// helper classes
// ----------------------------------------------------------------------------

/// structure to hold MailFolder pointer and associated mailstream pointer
class StreamConnection
{
public:
   /// pointer to a MailFolderCC object
   MailFolderCC    *folder;
   /// pointer to the associated MAILSTREAM
   MAILSTREAM   const *stream;
   /// name of the MailFolderCC object
   String name;
   /// for POP3/IMAP/NNTP: login or newsgroup
   String login;

   StreamConnection()
   {
      folder = NULL;
      stream = NULL;
   }

#ifdef DEBUG
   virtual ~StreamConnection();
#endif // DEBUG
};

KBLIST_DEFINE(StreamConnectionList, StreamConnection);
KBLIST_DEFINE(FolderViewList, FolderView);


/**
   MailFolder class, implemented with the C-client library.

   This class implements the functionality defined by the MailFolder
   parent class by using the c-client library for mailbox access. It
   handles all the callbacks from the library and routes them to the
   relevant objects by using a static list of objects and associated
   MAILSTREAM pointers.
*/
class MailFolderCC : public MailFolderCmn
{
public:
   /// flags for DoSetSequenceFlag()
   enum SequenceKind
   {
      SEQ_UID,
      SEQ_MSGNO
   };

   /** @name Constructors and destructor */
   //@{
   static MailFolderCC * OpenFolder(const MFolder *mfolder,
                                    OpenMode openmode,
                                    wxFrame *frame);

   static bool CloseFolder(const MFolder *mfolder);
   //@}

   static int CloseAll();
   static MailFolder **GetAllOpened();

   /** Phyically deletes this folder.
       @return true on success
   */
   static bool DeleteFolder(const MFolder *mfolder);

   static bool Rename(const MFolder *mfolder, const String& name);

   static long ClearFolder(const MFolder *folder);

   static bool CheckStatus(const MFolder *folder);

   /// return the directory of the newsspool:
   static String GetNewsSpool(void);

   /// wrapper for FindFolder() version below
   static MailFolderCC *FindFolder(const MFolder* mfolder);

   /// checks whether a folder with that path exists
   static MailFolderCC *FindFolder(const String& path, const String& login);

   /** Checks if it is OK to exit the application now.
        @param which Will either be set to empty or a '\n' delimited
        list of folders which are in critical sections.
   */
   static bool CanExit(String *which);

   virtual bool IsOpened(void) const { return m_MailStream != NULL; }

   /** return the full folder name
       @return the folder's name
   */
   virtual String GetName(void) const { return m_mfolder->GetFullName(); }

   /// return the folder type
   virtual MFolderType GetType(void) const { return m_mfolder->GetType(); }

   /// return the folder flags
   virtual int GetFlags(void) const { return m_mfolder->GetFlags(); }

   /// Return IMAP spec
   virtual String GetImapSpec(void) const { return m_ImapSpec; }

   /** Get the profile.
       @return Pointer to the profile.
   */
   virtual Profile *GetProfile(void) const { return m_Profile; }

   /// Checks if the folder is in a critical section.
   bool InCritical(void) const { return m_InCritical; }

   /// @name Folder operations
   //@{

   // count various kinds of messages
   virtual unsigned long CountNewMessages(void) const;
   virtual unsigned long CountRecentMessages(void) const;
   virtual unsigned long CountUnseenMessages(void) const;
   virtual unsigned long CountDeletedMessages(void) const;

   // uid -> msgno
   virtual MsgnoType GetMsgnoFromUID(UIdType uid) const;

   /** get message header
       @param uid mesage uid
       @return message header information class
   */
   virtual Message *GetMessage(unsigned long uid);

   /** Set flags on a sequence of messages. Possible flag values are MSG_STAT_xxx
       @param sequence the IMAP sequence
       @param flag flag to be set, e.g. "\\Deleted"
       @param set if true, set the flag, if false, clear it
       @return always true UNSUPPORTED!
   */
   virtual bool SetSequenceFlag(const String& sequence,
                                int flag,
                                bool set = true);

   virtual bool SetFlagForAll(int flag, bool set = true);

   /** Set flags on a sequence of messages. Possible flag values are MSG_STAT_xxx
       @param sequence the IMAP sequence of uids
       @param flag flag to be set, e.g. "\\Deleted"
       @param set if true, set the flag, if false, clear it
       @return true on success
   */
   virtual bool SetFlag(const UIdArray *sequence,
                        int flag,
                        bool set = true);

   /// override base class SaveMessages() to do server side copy if possible
   virtual bool SaveMessages(const UIdArray *selections, MFolder *folder);

   virtual bool AppendMessage(const Message & msg);

   virtual bool AppendMessage(const String& msg);
   virtual void ExpungeMessages(void);


   virtual MsgnoArray *SearchByFlag(MessageStatus flag,
                                    int flags = SEARCH_SET |
                                                SEARCH_UNDELETED,
                                    MsgnoType last = 0) const;

   virtual UIdArray *SearchMessages(const class SearchCriterium *crit);


   virtual bool ThreadMessages(const ThreadParams& thrParams,
                               ThreadData *thrData);

   virtual bool SortMessages(MsgnoType *msgnos, const SortParams& sortParams);


   /** Set a flag for all messages in this sequence which may contain either
       UIDs or msgnos
    */
   virtual bool DoSetSequenceFlag(SequenceKind kind,
                                  const String& sequence,
                                  int flag,
                                  bool set = true);

   /** Check whether mailbox has changed.
       @return FALSE on error
   */
   virtual bool Ping(void);

   /** Perform a checkpoint on the folder. */
   virtual void Checkpoint(void);

   //@}

   /**@name Subscription management */
   //@{
   /** Subscribe to a given mailbox (related to the
       mailfolder/mailstream underlying this folder.
       @param host the server host, or empty for local newsspool
       @param protocol MF_IMAP or MF_NNTP or MF_NEWS
       @param mailboxname name of the mailbox to subscribe to
       @param bool if true, subscribe, else unsubscribe
       @return true on success
   */
   static  bool Subscribe(const String &host,
                          MFolderType protocol,
                          const String &mailboxname,
                          bool subscribe = true);

   /** Get a listing of all mailboxes.

       DO NOT USE THIS FUNCTION, BUT ASMailFolder::ListFolders instead!!!

       @param asmf the ASMailFolder initiating the request
       @param pattern a wildcard matching the folders to list
       @param subscribed_only if true, only the subscribed ones
       @param reference implementation dependend reference
    */
   void ListFolders(class ASMailFolder *asmf,
                    const String &pattern = "*",
                    bool subscribed_only = false,
                    const String &reference = "",
                    UserData ud = 0,
                    Ticket ticket = ILLEGAL_TICKET);

   //@}
   /**@name Access control */
   //@{
   /** Locks a mailfolder for exclusive access. In multi-threaded mode
       it really locks it and always returns true. In single-threaded
       mode it returns false if we cannot get a lock.
       @return TRUE if we have the lock
   */
   virtual bool Lock(void) const;
   /** Releases the lock on the mailfolder. */
   virtual void UnLock(void) const;
   /// Is folder locked?
   virtual bool IsLocked(void) const;
   //@}

   /**@name Functions to get an overview of messages in the folder. */
   //@{
   virtual unsigned long GetMessageCount() const;

   virtual MsgnoType GetHeaderInfo(ArrayHeaderInfo& headers,
                                   const Sequence& seq);
   //@}

   virtual char GetFolderDelimiter() const;

   /// return TRUE if CClient lib had been initialized
   static bool IsInitialized() { return ms_CClientInitialisedFlag; }

   /**
      check whether a folder can have subfolders
   */
   static bool HasInferiors(const String &imapSpec,
                            const String &user,
                            const String &pw);

   /** @name Functions used by MessageCC for read progress. */
   //@{
   /**
     Prepare for reading the given amount of data from folder.

     @param total the amount of bytes to be read
    */
   void StartReading(unsigned long total);

   /**
     Called when data is read
    */
   void EndReading();
   //@}

   /**
      For use by class MessageCC only

      @return MAILSTREAM of the folder
   */
   MAILSTREAM *Stream(void) const { return m_MailStream; }

   /**
     Methods for use by CCEventReflector only: see the comment in
     MailFolderCC.cpp for more details.
   */
   //@{

   /// called to process new mail in the folder (filter it, ...)
   void OnNewMail();

   /// called to generate MsgStatus event for many messages at once
   void OnMsgStatusChanged();

   //@}

private:
   /** @name Constructors and such */
   //@{
   /// private constructor, does basic initialisation
   MailFolderCC(const MFolder *mfolder, wxFrame *frame);

   /// Common code for constructors
   void Create(MFolderType type, int flags);

   /// code common to Create() and Close()
   void Init();

   /// destructor closes the folder automatically
   ~MailFolderCC();
   //@}

   /** @name Opening/closing the folders */
   //@{
   /** Try to open the mailstream for this folder.
       @return true on success
   */
   bool Open(OpenMode openmode = Normal);

   /// physically create the file (MF_FILE or MF_MH) folder
   void CreateFileFolder();

   /// check (and delete if requested) for the lock on this file folder
   bool CheckForFileLock();

   /// Close the folder
   virtual void Close(void);

   //@}

   /** @name Mailbox update helpers */
   //@{

   /// call to notify everybody that its listing changed
   virtual void RequestUpdate(void);

   /// call to notify everybody that some messages were expunged
   void RequestUpdateAfterExpunge();

   //@}

   /** @name Message counting */
   //@{
   /// perform the search, return number of messages found
   unsigned long SearchAndCountResults(struct search_program *pgm) const;

   /// helper of SearchAndCountResults() and SearchByFlag()
   MsgnoArray *DoSearch(struct search_program *pgm, int ccSearchFlags) const;

   /// called by CountAllMessages() to perform actual counting
   virtual bool DoCountMessages(MailFolderStatus *status) const;
   //@}

   /// Update the timeout values from a profile
   void UpdateTimeoutValues(void);

   /// apply all timeout values
   void ApplyTimeoutValues(void);

   /// helper of OverviewHeader
   static String ParseAddress(struct mail_address *adr);

   /**
     Called from GetHeaderInfo() to process one header

     @return false to abort overview generation, true to continue.
   */
   bool OverviewHeaderEntry(class OverviewData *overviewData,
                            struct message_cache *elt,
                            struct mail_envelope *env);

   /** We remember the last folder to enter a critical section, helps
       to find crashes.*/
   static String ms_LastCriticalFolder;

   /// Build the sequence string from the array of message uids
   static String BuildSequence(const UIdArray& messages);

   /**
      Try to create folder if it hadn't been created yet, returns true if the
      folder could be created and opened successfully or false if the folder
      had to be created but its creation failed.

      @param folder to create/open
      @param pStream will be filled with a valid stream if not NULL and
                     if we indeed created the folder (and will be NULL if we
                     didn't do anything)
      @return false if the folder couldn't be created
   */
   static bool CreateIfNeeded(const MFolder *folder,
                              MAILSTREAM **pStream = NULL);

   /// just do mail_ping() on the opened folder, return TRUE if ok
   bool PingOpenedFolder();

   /**
     Called when the folder was closed not in result of calling our Close():
     this will reset m_MailStream to NIL, call Close() to clean up and notyf
     the outside world about our closure
   */
   void ForceClose();

   /// Updates the status of a single message.
   void UpdateMessageStatus(unsigned long seqno);

   /// update the folder after appending messages to it
   void UpdateAfterAppend();

   virtual void ReadConfig(MailFolderCmn::MFCmnOptions& config);

   /// clear m_expungedMsgnos and m_expungedPositions arrays
   void DiscardExpungeData();

   /** @name Notification handlers */
   //@{
   /// called when the number of the messages in the folder changes
   void OnMailExists(struct mail_stream *stream, MsgnoType msgnoMax);

   /// called when the given msgno is expunged from the folder
   void OnMailExpunge(MsgnoType msgno);
   //@}

   // members (mostly) from here on
   // -----------------------------

   /** @name Authentification info */
   //@{

   /// The following is also called by SendMessageCC for ESMTP authentication
   static void SetLoginData(const String &user, const String &pw);

   /// for POP/IMAP boxes, this holds the user id for the callback
   static String MF_user;

   /// for POP/IMAP boxes, this holds the password for the callback
   static String MF_pwd;

   /// do we have login/password?
   bool HasLogin() const { return !m_login.empty() && !m_password.empty(); }

   /// do we need to authentificate?
   bool NeedsAuthInfo() const
      { return !HasLogin() && m_mfolder->NeedsLogin(); }

   /// propose the user to save the password temporarily or permanently
   void ProposeSavePassword();

   /// temporarily (i.e. for this session only) save password if we have it
   void SavePasswordForSession();

   /**
     Try to find login/password for the given folder looking among the
     logins/passwords previosuly entered by user if they are not stored in the
     folder itself. If it doesn't find it there neither, asks the user for the
     password (and sets didAsk to true then)

     @param mfolder the folder we need auth info for
     @param login the variable containing username
     @param password the variable containing password
     @param didAsk a pointer (may be NULL) set to true if password was entered
     @return true if the password is either not needed or was entered
    */
   static bool GetAuthInfoForFolder(const MFolder *mfolder,
                                    String& login,
                                    String& password,
                                    bool *didAsk = NULL);
   //@}

   /** @name c-client initialization */
   //@{
   /// has c-client library been initialised?
   static bool ms_CClientInitialisedFlag;

   /// initialise c-client library
   static void CClientInit(void);
   //@}

   /** @name Mail folder parameters */
   //@{

   /// the object containing all our parameters
   MFolder *m_mfolder;

   /// Full folder spec for c-client
   String m_ImapSpec;

   /// the folder name delimiter, (char)-1 if unknown yet
   char m_chDelimiter;

   /// the profile we use for our options
   Profile *m_Profile;

   /**
     Note that login/password may be different from m_mfolder data (i.e. what
     GetLogin()/Password() return) if the user had chosen to not store them in
     the config but enters them interactively.
    */

   /// username to use for opening this folder
   String m_login;

   /// password to use for opening this folder
   String m_password;

   //@}

   /** @name Mail folder state */
   //@{

   /// mailstream associated with this folder
   MAILSTREAM *m_MailStream;

   /// the number of messages as we know it
   MsgnoType m_nMessages;

   /// last seen UID
   UIdType m_uidLast;

   /// last seen UID of a new message
   UIdType m_uidLastNew;

   /// UID validity (in IMAP/c-client sense) for this folder
   UIdType m_uidValidity;

   //@}

   /** @name Temporary operation parameters */
   //@{

   /**
     This flag is set to true during the window between mm_exists()
     notification and the moment we send an update request to the GUI (which
     happens during the next idle time iteration normally).

     During this time span the listing (m_headers) is in an intermediate state
     as its internal msgno <-> position mapping may be out of date and it
     should be used with extreme care to avoid possible c-client reentrancies!
    */
   bool m_gotUnprocessedNewMail;

   /**
     These two arrays are used between the moment when we get the expunge
     notification and until the moment we can send the notification about it
     to the GUI.

     We need both msgnos and positions because by the time GUI code gets our
     notification, the header listing doesn't have the items corresponding to
     the expunged msgnos (they were expunged!) and so can't be asked for the
     positions of these msgnos but the GUI needs them.
    */
   //@{

   /// the array containing indices of expunged messages or NULL
   wxArrayInt *m_expungedMsgnos;

   /// the array containing the positions of expunged messages or NULL
   wxArrayInt *m_expungedPositions;

   //@}

   /**
     The elements are added to these arrays from mm_flags() handler
     (i.e. UpdateMessageStatus()) and they are cleared later in
     OnMsgStatusChanged()
    */
   //@{

   /// the array containing the msgnos of the messages whose status changed
   wxArrayInt *m_statusChangedMsgnos;

   /// the arrays containing the old and new status of these messages
   wxArrayInt *m_statusChangedOld,
              *m_statusChangedNew;

   //@}

   /** If we are searching, this points to an UIdArray where to store
       the entries found.
   */
   UIdArray *m_SearchMessagesFound;

   /**@name only used for ListFolders: */
   //@{
   UserData m_UserData;
   Ticket   m_Ticket;
   ASMailFolder *m_ASMailFolder;
   //@}

   /// Used by the subscription management.
   class FolderListingCC *m_FolderListing;

   //@}

   /** @name Locking and other semaphores */
   //@{

   /// Is this folder in a critical c-client section?
   bool m_InCritical;

   /// when this mutex is locked, this folder can't be accessed at all
   MMutex m_mutexExclLock;

   /// locked while we're processing the new mail which just arrived
   MMutex m_mutexNewMail;

   //@}

   /** @name functions for mapping mailstreams and objects
       These functions enable the class to map incoming events from
       the  c-client library to the object associated with that event.
       */
   //@{

   /// a pointer to the object to use as default if lookup fails
   static MailFolderCC *ms_StreamListDefaultObj;

   /// mapping MAILSTREAM* to objects of this class and their names
   static StreamConnectionList ms_StreamList;

   /// adds this object to Map
   void AddToMap(const MAILSTREAM *stream, const String& login) const;

   /// remove this object from Map
   void RemoveFromMap(void) const;

   /// find the stream in the map
   static MailFolderCC *LookupStream(const MAILSTREAM *stream);

   /// lookup object in map first by stream, then by name
   static MailFolderCC *LookupObject(const MAILSTREAM *stream,
                                     const char *name = NULL);
   //@}

   /** @name c-client parameters */
   //@{
   /// IMAP lookahead value
   int m_LookAhead;
   /// TCP/IP open timeout in seconds.
   int m_TcpOpenTimeout;
   /// TCP/IP read timeout in seconds.
   int m_TcpReadTimeout;
   /// TCP/IP write timeout in seconds.
   int m_TcpWriteTimeout;
   /// TCP/IP close timeout in seconds.
   int m_TcpCloseTimeout;
   /// rsh connection timeout in seconds.
   int m_TcpRshTimeout;
   /// ssh connection timeout in seconds.
   int m_TcpSshTimeout;
   /// the path to rsh
   String m_RshPath;
   /// the path to ssh
   String m_SshPath;
   //@}

#ifdef DEBUG
   /// print a list of all streams
   static void DebugStreams(void);
#endif

   friend bool MailFolderCCInit();
   friend void MailFolderCCCleanup();

   friend class CCDefaultFolder;
   friend class SendMessageCC;

public:
   DEBUG_DEF
   MOBJECT_DEBUG(MailFolderCC)

public:
   /** @name common callback routines
       They all take a stram argument and the number of a message.
       Do not call them, they are only for use by the c-client library!
   */
   //@{

   /// this message matches a search
   static void mm_searched(MAILSTREAM *stream, unsigned long number);

   /// tells program that there are this many messages in the mailbox
   static void mm_exists(MAILSTREAM *stream, unsigned long number);

   /** deliver stream message event
       @param stream mailstream
       @param str message str
       @param errflg error level
       */
   static void mm_notify(MAILSTREAM *stream, String str, long
                         errflg);

   /** this mailbox name matches a listing request
       @param stream mailstream
       @param delim   character that separates hierarchies
       @param name    mailbox name
       @param attrib   mailbox attributes
       */
   static void mm_list(MAILSTREAM *stream, char delim, String name,
                       long attrib);

   /** matches a subscribed mailbox listing request
       @param stream   mailstream
       @param delim   character that separates hierarchies
       @param name   mailbox name
       @param attrib   mailbox attributes
       */
   static void mm_lsub(MAILSTREAM *stream, char delim, String name,
                       long attrib);
   /** status of mailbox has changed
       @param stream   mailstream
       @param mailbox    mailbox name for this status
       @param status   structure with new mailbox status
       */
   static void mm_status(MAILSTREAM *stream, String mailbox, MAILSTATUS *status);

   /** log a message
       @param str   message string
       @param errflg   error level
       @param mf if non-NULL the folder
       */
   static void mm_log(const String& str, long errflg, MailFolderCC *mf = NULL);

   /** log a debugging message
       @param str    message string
       */
   static void mm_dlog(const String& str);

   /** get user name and password
       @param   mb   parsed mailbox specification
       @param   user    pointer where to return username
       @param   pwd   pointer where to return password
       @param    trial   number of prior login attempts
       */
   static void mm_login(NETMBX *mb, char *user, char *pwd, long trial);

   /** alert that c-client will run critical code
       @param   stream   mailstream
   */
   static void mm_critical(MAILSTREAM *stream);

   /**   no longer running critical code
   @param   stream mailstream
     */
   static void mm_nocritical(MAILSTREAM *stream);

   /** unrecoverable write error on mail file
       @param   stream   mailstream
       @param   errcode   OS error code
       @param   serious   non-zero: c-client cannot undo error
       @return   abort flag: if serious error and abort non-zero: abort, else retry
       */
   static long mm_diskerror(MAILSTREAM *stream, long errcode, long serious);

   /** program is about to crash!
       @param   str   message string
       */
   static void mm_fatal(char *str);

   /// gets called when messages were deleted
   static void mm_expunged(MAILSTREAM *stream, unsigned long number);

   /// gets called when flags for a message have changed
   static void mm_flags(MAILSTREAM *stream, unsigned long number);
   //@}
};

#endif // MAILFOLDERCC_H

